import unittest
import numpy as np
import numpsy as nsy

class TestHelpers(unittest.TestCase):

    def test_select_available_property_(self):
        self.assertEqual(nsy.helpers.__select_available_property__("first", "default", "default"), "first")
        self.assertEqual(nsy.helpers.__select_available_property__("default", "second", "default"), "second")
        self.assertEqual(nsy.helpers.__select_available_property__("default",
                                                                   "default",
                                                                   "default",
                                                                   "optional"),
                         "optional")
        self.assertEqual(nsy.helpers.__select_available_property__("default",
                                                                   "default",
                                                                   "default",
                                                                   "default",
                                                                   "optional"
                                                                   ),
                         "optional")
        self.assertEqual(nsy.helpers.__select_available_property__(nsy.configuration.undefined_unit_symbol,
                                                                   nsy.configuration.undefined_unit_symbol,
                                                                   nsy.configuration.undefined_unit_symbol),
                         nsy.configuration.undefined_unit_symbol)
        self.assertEqual(nsy.helpers.__select_available_property__(nsy.configuration.undefined_unit_symbol,
                                                                   nsy.configuration.undefined_unit_symbol,
                                                                   nsy.configuration.undefined_unit_symbol,
                                                                   nsy.configuration.undefined_unit_symbol),
                         nsy.configuration.undefined_unit_symbol)
        self.assertEqual(nsy.helpers.__select_available_property__("",
                                                                   "",
                                                                   "default"),
                         "default")


class TestConfiguration(unittest.TestCase):

    def test_defaults(self):
        self.assertEqual("Ø", nsy.configuration.undefined_value_symbol)
        # TODO

class TestCore(unittest.TestCase):

    def setUp(self):
        self.compounded_constant_1 = nsy.C(s="k", na="Boltzmann constant", n=1.380649e-23,
                                           unit=nsy.U(s="J") / nsy.U(s="K"))
        self.compounded_variable_1 = nsy.V(s="T_a", na="Ambient temperature", n=np.linspace(0.1, 300),
                                           unit=nsy.U(s="K"))
        self.compounded_variable_2 = nsy.V(s="q", na="single electron charge", n=1.602176634e-19, unit=nsy.U(s="J"))

    def test_default_initialization(self):
        self.unit = nsy.Unit()
        self.assertEqual(self.unit.__name__, nsy.configuration.undefined_unit_name)
        self.assertEqual(self.unit.__name_expression__, nsy.configuration.undefined_unit_name)
        self.assertEqual(self.unit.__symbol__, nsy.configuration.undefined_unit_symbol)
        self.assertEqual(self.unit.__symbolic_expression__, nsy.configuration.undefined_unit_symbolic_expression)

    def test_shortcut_initialization(self):
        self.unit_shortcut = nsy.Unit(na="unit", s="u")
        self.unit_long = nsy.Unit(name="unit", symbol="u")
        self.assertEqual(self.unit_shortcut.name, self.unit_long.name)
        self.assertEqual(self.unit_shortcut.na, self.unit_long.na)
        self.assertEqual(self.unit_shortcut.symbol, self.unit_long.symbol)
        self.assertEqual(self.unit_shortcut.s, self.unit_long.s)

    def test_value_initialization_integrated_units(self):
        self.assertIsNotNone(self.compounded_constant_1)
        self.assertIsNotNone(self.compounded_variable_1)
        self.assertIsNotNone(self.compounded_variable_2)

    def test_multiplication(self):
        self.variable_shorthand_1 = nsy.V(s="\mu_{nb}", n=1)
        self.variable_shorthand_2 = nsy.V(s="\mu_{pe}", n=1)
        self.assertIsNotNone(self.variable_shorthand_1 * self.variable_shorthand_2)

    def test_complex_multiplication_exp_sqrt(self):
        self.assertTrue(nsy.sqrt(self.compounded_variable_2) * nsy.e(self.compounded_variable_1))

class TestFunctions(unittest.TestCase):
    def test_exp(self):
        self.assertTrue(nsy.exp(1))
        self.assertTrue(nsy.log(1))
        self.assertTrue(nsy.log10(1))

if __name__ == '__main__':
    unittest.main()



import numpy as np
import sympy as sy
from collections.abc import Iterable
from . import configuration
from . import core


def __check_properties__(instance, property):
    if hasattr(instance.properties, property):
        out = instance.properties[property]
    else:
        out = None


def __validate_numeric__(input):
    # TODO implement
    return input


def __unit_variable_generator__(instance):
    if hasattr(instance, "unit"):
        if instance.unit == "":
            # TODO in case we want show other default
            instance_unit_variable = instance.unit
        else:
            instance_unit_variable = instance.unit
    else:
        instance_unit_variable = core.undefined_unit_default
    return instance_unit_variable


def __unit_variables_generator__(first, second):
    first_unit_variable = __unit_variable_generator__(first)
    second_unit_variable = __unit_variable_generator__(second)
    return [first_unit_variable, second_unit_variable]


def name_variable_generator(instance):
    if hasattr(instance, "name") and (instance.name != ""):
        if not isinstance(instance.name, str):
            instance_name_variable = str(instance.name)
        else:
            instance_name_variable = instance.name
    else:
        if hasattr(instance, "name_expression"):
            if not isinstance(instance.name_expression, str):
                instance_name_variable = str(instance.name_expression)
            else:
                instance_name_variable = instance.name_expression
        else:
            instance_name_variable = str(instance)
    return instance_name_variable


def name_variables_generator(first, second):
    first_name_variable = name_variable_generator(first)
    second_name_variable = name_variable_generator(second)
    return [first_name_variable, second_name_variable]


def numerical_variable_generator(instance):
    if hasattr(instance, "numerical"):
        instance_numerical = instance.numerical
    else:
        instance_numerical = instance
    return instance_numerical


def numerical_variables_generator(first, second):
    first_numerical = numerical_variable_generator(first)
    second_numerical = numerical_variable_generator(second)
    return [first_numerical, second_numerical]


def symbolic_expression_variable_generator(instance):
    if hasattr(instance, "__symbol__"):
        if instance.__symbol__ == configuration.undefined_unit_symbol:
            if hasattr(instance, "symbolic_expression"):
                if isinstance(instance.__symbolic_expression__, Iterable):
                    if instance.__symbolic_expression__[0] == configuration.undefined_value_symbolic_expression:
                        if hasattr(instance, "numerical"):
                            instance_symbolic_variable = instance.numerical
                        else:
                            instance_symbolic_variable = sy.Symbol(configuration.undefined_unit_symbol)
                    else:
                        instance_symbolic_variable = instance.symbolic_expression[0]
                else:
                    if instance.__symbolic_expression__ == configuration.undefined_value_symbolic_expression:
                        if hasattr(instance, "numerical"):
                            instance_symbolic_variable = instance.numerical
                        else:
                            instance_symbolic_variable = sy.Symbol(configuration.undefined_unit_symbol)
                    else:
                        instance_symbolic_variable = instance.symbolic_expression
            elif hasattr(instance, "numerical"):
                instance_symbolic_variable = instance.numerical
            else:
                instance_symbolic_variable = sy.Symbol(configuration.undefined_unit_symbol)
        else:
            instance_symbolic_variable = instance.symbol
    # TODO Should this handle numpy arrays?
    elif np.isscalar(instance):
        instance_symbolic_variable = instance
    else:
        instance_symbolic_variable = sy.Symbol(configuration.undefined_unit_symbol)
    return instance_symbolic_variable


def symbolic_expression_variables_generator(first, second):
    first_symbolic_variable = symbolic_expression_variable_generator(first)
    second_symbolic_variable = symbolic_expression_variable_generator(second)
    return [first_symbolic_variable, second_symbolic_variable]


def unit_variable_generator(instance):
    if hasattr(instance, "unit"):
        return instance.unit
    else:
        return core.Unit()


def full_variable_generator(instance):
    if configuration.setup.calculation_style == "numpsy":
        name_variable = name_variable_generator(instance)
        symbolic_variable = symbolic_expression_variable_generator(instance)
        numerical_variable = numerical_variable_generator(instance)
        unit_variable = unit_variable_generator(instance)
        return {
            "name": name_variable,
            "symbolic": symbolic_variable,
            "numerical": numerical_variable,
            "unit": unit_variable,
        }
    if configuration.setup.calculation_style == "numpy":
        # name_variable = name_variable_generator(instance)
        # symbolic_variable = symbolic_expression_variable_generator(instance)
        numerical_variable = numerical_variable_generator(instance)
        # unit_variable = unit_variable_generator(instance)
        return {
            # "name": name_variable,
            # "symbolic": symbolic_variable,
            "numerical": numerical_variable,
            # "unit": unit_variable,
        }


def __select_available_property__(first, second, default, *kwargs):
    # TODO how robust is this fix?
    if type(first) == core.Unit:
        first_str = str(symbolic_expression_variable_generator(first))
    else:
        first_str = str(first)

    if type(second) == core.Unit:
        second_str = str(symbolic_expression_variable_generator(second))
    else:
        second_str = str(second)

    if isinstance(default, list):
        default_i_str = str(default[0])
        default_i = default[0]
    else:
        if type(default) == core.Unit:
            default_i_str = str(symbolic_expression_variable_generator(default))
            default_i = default
        else:
            default_i_str = str(default)
            default_i = default

    if (first_str != default_i_str) and (second_str == default_i_str) and bool(first_str):
        return first
    elif (first_str == default_i_str) and (second_str != default_i_str) and bool(second_str):
        return second
    elif (first_str == second_str) and (first_str != default_i_str) and (second_str != default_i_str) and bool(
            first_str):
        return first
    elif (first_str == default_i_str) and (second_str == default_i_str):
        try:
            if bool(kwargs):
                for kwarg in kwargs:
                    if isinstance(default, list):
                        if bool(kwarg) and all((kwarg != str(default_j)) for default_j in default):
                            return kwarg
                    else:
                        if (kwarg != default_i_str):
                            return kwarg
                return default_i
            else:
                return default_i
        except:
            return default_i
    elif (first_str != second_str) and (first_str != default_i_str) and (second_str != default_i_str):
        # TODO Workaround of undiagnosed problem where default_i_str gets a strange default.
        if (first_str != None):
            if configuration.setup.debug_mode == True:
                print(ValueError("Incongruent properties assignments, assigned first: "
                                 + str(first_str)
                                 + ", when second_str: "
                                 + str(second_str) + " and default: " + str(default_i_str)))
            return first
        else:
            if configuration.setup.debug_mode == True:
                raise ValueError("Incongruent properties assignments, first: "
                                     + str(first_str)
                                     + ", second_str: "
                                     + str(second_str) + " default: " + str(default_i_str))
    else:
        if configuration.setup.debug_mode == True:
            print(Warning("Incompatible property assignment, first: " + str(first) + ", second_str: " + str(
                second) + ", assigning default: " + str(default_i)))
        return default_i



import sympy as sy
import pandas as pd
from . import configuration
from . import helpers
from . import printers


class DataMixin:
    def __init__(self):
        super(DataMixin, self).__init__()

    @property
    def data(self):
        if configuration.setup.calculation_style == "numpsy":
            data = {}
            if hasattr(self, "name"):
                data["name"] = self.name
            if hasattr(self, "name_expression"):
                data["name_expression"] = self.name_expression
            if hasattr(self, "note"):
                data["note"] = self.note
            if hasattr(self, "numerical"):
                data["numerical"] = self.numerical
            if hasattr(self, "symbol"):
                if bool(self.symbol):
                    a = sy.latex(self.symbol, mode="equation")
                    data["symbol"] = a
                else:
                    raise ValueError(
                        "Sympy incompatible symbol input: self.symbol: " + str(self.__symbol__) + " for class: " + str(
                            self.__class__))
            if hasattr(self, "symbolic_expression"):
                if sy.latex(self.symbolic_expression):
                    b = sy.latex(self.symbolic_expression, mode="equation")
                else:
                    b = ""
                data["symbolic_expression"] = b
            if hasattr(self, "unit"):
                # if self.unit.symbol:
                #    data["unit"] = "$" + self.unit.__symbol__ + "$"
                # else:
                #    data["unit"] = "$" + sy.latex(self.unit.data.symbolic_expression) + "$"
                self.unit.data.symbol = "$" + self.unit.data.symbol + "$"
                self.unit.data.symbolic_expression = (
                        "$" + self.unit.data.symbolic_expression + "$"
                )
                data["unit"] = (
                        "Symbol: "
                        + self.unit.data["symbol"].values[0]
                        + "\n"
                        + "Symbolic Expression: "
                        + self.unit.data["symbolic_expression"].values[0]
                )
            return pd.DataFrame([data], index=[self.__class__.__name__])
        if configuration.setup.calculation_style == "numpy":
            data = {}
            # if hasattr(self, "name"):
            #     data["name"] = self.name
            # if hasattr(self, "name_expression"):
            #     data["name_expression"] = self.name_expression
            # if hasattr(self, "note"):
            #     data["note"] = self.note
            if hasattr(self, "numerical"):
                data["numerical"] = self.numerical
            # if hasattr(self, "symbol"):
            #     if bool(self.symbol):
            #         a = sy.latex(self.symbol, mode="equation")
            #         data["symbol"] = a
            #     else:
            #         raise ValueError(
            #             "Sympy incompatible symbol input: self.symbol: " + str(self.__symbol__) + " for class: " + str(
            #                 self.__class__))
            # if hasattr(self, "symbolic_expression"):
            #     if sy.latex(self.symbolic_expression):
            #         b = sy.latex(self.symbolic_expression, mode="equation")
            #     else:
            #         b = ""
            #     data["symbolic_expression"] = b
            # if hasattr(self, "unit"):
            #     # if self.unit.symbol:
            #     #    data["unit"] = "$" + self.unit.__symbol__ + "$"
            #     # else:
            #     #    data["unit"] = "$" + sy.latex(self.unit.data.symbolic_expression) + "$"
            #     self.unit.data.symbol = "$" + self.unit.data.symbol + "$"
            #     self.unit.data.symbolic_expression = (
            #             "$" + self.unit.data.symbolic_expression + "$"
            #     )
            #     data["unit"] = (
            #             "Symbol: "
            #             + self.unit.data["symbol"].values[0]
            #             + "\n"
            #             + "Symbolic Expression: "
            #             + self.unit.data["symbolic_expression"].values[0]
            #     )
            return pd.DataFrame([data], index=[self.__class__.__name__])

    @data.setter
    def data(self, value):
        if hasattr(self, "name") & hasattr(value, "name"):
            self.name = value.name
        if hasattr(self, "name_expression") & hasattr(value, "name_expression"):
            self.name_expression = value.name_expression
        if hasattr(self, "note") & hasattr(value, "note"):
            self.note = value.note
        if hasattr(self, "numerical") & hasattr(value, "numerical"):
            self.numerical = value.numerical
        if hasattr(self, "symbol") & hasattr(value, "symbol"):
            self.symbol = value.__symbol__
        if hasattr(self, "symbolic_expression") & hasattr(value, "symbolic_expression"):
            self.symbolic_expression = value.__symbolic_expression__
        if hasattr(self, "unit") & hasattr(value, "unit"):
            self.unit = value.unit


class InstanceMixin(DataMixin):
    def __init__(
            self,
            name=configuration.undefined_unit_name,
            n=configuration.undefined_unit_name,
            name_expression=configuration.undefined_unit_name,
            note=configuration.undefined_note,
            print_style=configuration.undefined_print_style
    ):
        super(InstanceMixin, self).__init__()
        if configuration.setup.calculation_style == "numpsy":
            self.__name__ = helpers.__select_available_property__(name,
                                                                  n,
                                                                  configuration.undefined_unit_name)
            self.__name_expression__ = helpers.__select_available_property__(name_expression,
                                                                             self.__name__,
                                                                             configuration.undefined_unit_name)
            self.__note__ = note
            self.__print_style__ = print_style
            self.data
        if configuration.setup.calculation_style == "numpy":
            # self.__name__ = helpers.__select_available_property__(name,
            #                                                       n,
            #                                                       configuration.undefined_unit_name)
            # self.__name_expression__ = helpers.__select_available_property__(name_expression,
            #                                                                  self.__name__,
            #                                                                  configuration.undefined_unit_name)
            # self.__note__ = note
            self.__print_style__ = print_style
            self.data

    @property
    def name(self):
        """Return name string"""
        return self.__name__

    @name.setter
    def name(self, value):
        self.__name__ = value

    @property
    def name_expression(self):
        """Return name expression consisting of an equivalent string representation to name"""
        return self.__name_expression__

    @name_expression.setter
    def name_expression(self, value):
        """Set name expression consisting of an equivalent string representation to name."""
        self.__name_expression__ = value

    @property
    def note(self):
        """Return note string"""
        return self.__note__

    @note.setter
    def note(self, value):
        self.__note__ = value

    @property
    def print_style(self):
        """Return name string"""
        return self.__print_style__

    @print_style.setter
    def print_style(self, value):
        self.__print_style__ = value

    def __repr__(self):
        return printers.__repr__(self)

    def _repr_markdown_(self):
        return printers._repr_markdown_(self)

    na = name
    no = note
    nae = name_expression
    p = print_style


class Unit(InstanceMixin):
    def __init__(
            self,
            name=configuration.undefined_unit_name,
            na=configuration.undefined_unit_name,
            name_expression=configuration.undefined_unit_name,
            note=configuration.undefined_note,
            symbol=configuration.undefined_unit_symbol,
            s=configuration.undefined_unit_symbol,
            symbolic_expression=sy.Symbol(configuration.undefined_unit_symbol),
            print_style=configuration.undefined_print_style
    ):
        super(Unit, self).__init__()
        self.__name__ = helpers.__select_available_property__(name,
                                                              na,
                                                              [configuration.undefined_unit_name,
                                                               configuration.undefined_unit_symbol],
                                                              symbol,
                                                              s)
        self.__name_expression__ = helpers.__select_available_property__(name_expression,
                                                                         self.__name__,
                                                                         configuration.undefined_unit_name)
        self.__note__ = note
        self.__symbol__ = helpers.__select_available_property__(symbol,
                                                                s,
                                                                [configuration.undefined_unit_symbol,
                                                                 configuration.undefined_unit_name],
                                                                name)
        self.__symbolic_expression__ = helpers.__select_available_property__(symbolic_expression,
                                                                             self.symbol,
                                                                             configuration.undefined_unit_symbolic_expression)
        self.__print_style__ = print_style
        self.data

    def __add__(self, other):
        new = Unit()
        name_variables = helpers.name_variables_generator(self, other)
        new.name = "(" + name_variables[0] + "_plus_" + name_variables[1] + ")"
        symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
        new.symbolic_expression = symbol_variables[0] + symbol_variables[1]
        return new

    def __neg__(self):
        new = Unit()
        name_variable = helpers.name_variable_generator(self)
        new.name = (
                "minus" + name_variable
        )
        symbol_variable = helpers.symbolic_expression_variable_generator(self)
        new.symbolic_expression = - symbol_variable
        return new

    def __mul__(self, other):
        new = Unit()
        name_variables = helpers.name_variables_generator(self, other)
        new.name = "(" + name_variables[0] + "_times_" + name_variables[1] + ")"
        symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
        new.symbolic_expression = symbol_variables[0] * symbol_variables[1]
        return new

    def __sub__(self, other):
        new = Unit()
        name_variables = helpers.name_variables_generator(self, other)
        new.name = "(" + name_variables[0] + "_minus_" + name_variables[1] + ")"
        symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
        new.symbolic_expression = symbol_variables[0] - symbol_variables[1]
        return new

    def __pow__(self, other):
        new = Unit()
        name_variables = helpers.name_variables_generator(self, other)
        new.name = "(" + name_variables[0] + "_power_" + name_variables[1] + ")"
        symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
        new.symbolic_expression = symbol_variables[0] ** symbol_variables[1]
        return new

    def __truediv__(self, other):
        new = Unit()
        name_variables = helpers.name_variables_generator(self, other)
        new.name = "(" + name_variables[0] + ")_per_(" + name_variables[1] + ")"
        symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
        new.symbolic_expression = symbol_variables[0] / symbol_variables[1]
        return new

    @property
    def symbol(self):
        """Return unit symbol shorthand from Sympy"""
        if bool(self.__symbol__):
            return sy.Symbol(self.__symbol__)
        else:
            raise ValueError(
                "Sympy incompatible symbol input: self.symbol" + str(self.__symbol__) + "for class: " + str(
                    self.__class__))

    @symbol.setter
    def symbol(self, value):
        self.__symbol__ = value

    @property
    def symbolic_expression(self):
        """Return symbolic expression shorthand shorthand"""
        return self.__symbolic_expression__

    @symbolic_expression.setter
    def symbolic_expression(self, value):
        """Set symbolic expression shorthand shorthand"""
        self.__symbolic_expression__ = value

    s = symbol
    se = symbolic_expression


def undefined_unit_default_generator():
    return Unit(name=configuration.undefined_unit_name,
                symbol=configuration.undefined_unit_symbol,
                symbolic_expression=configuration.undefined_unit_symbolic_expression,
                )


undefined_unit_default = undefined_unit_default_generator()


class Value(InstanceMixin):
    @property
    def __parent_class__(self):
        return "Value"

    def __init__(
            self,
            name=configuration.undefined_value_name,
            na=configuration.undefined_value_name,
            name_expression=configuration.undefined_value_name,
            n=configuration.undefined_value_numerical,
            note=configuration.undefined_note,
            numerical=configuration.undefined_value_numerical,
            symbol=configuration.undefined_unit_symbol,
            s=configuration.undefined_unit_symbol,
            symbolic_expression=configuration.undefined_value_symbolic_expression,
            se=configuration.undefined_value_symbolic_expression,
            print_style=configuration.undefined_print_style,
            unit=Unit(),
            u=Unit(),
    ):
        super(Value, self).__init__()
        if configuration.setup.calculation_style == "numpsy":
            self.__name__ = helpers.__select_available_property__(name,
                                                                  na,
                                                                  [configuration.undefined_value_name,
                                                                   configuration.undefined_value_symbol],
                                                                  symbol,
                                                                  s
                                                                  )
            self.__name_expression__ = helpers.__select_available_property__(name_expression,
                                                                             self.__name__,
                                                                             configuration.undefined_value_name)
            self.__note__ = note
            self.__symbol__ = helpers.__select_available_property__(symbol,
                                                                    s,
                                                                    configuration.undefined_value_symbol)
            self.__symbolic_expression__ = helpers.__select_available_property__(symbolic_expression,
                                                                                 se,
                                                                                 configuration.undefined_value_symbolic_expression,
                                                                                 )
            self.__numerical__ = helpers.__validate_numeric__(helpers.__select_available_property__(
                numerical,
                n,
                configuration.undefined_value_numerical))
            self.__print_style__ = print_style
            self.__unit__ = helpers.__select_available_property__(first=unit,
                                                                  second=u,
                                                                  default=undefined_unit_default,
                                                                  )
        if configuration.setup.calculation_style == "numpy":
            # self.__name__ = helpers.__select_available_property__(name,
            #                                                       na,
            #                                                       [configuration.undefined_value_name,
            #                                                        configuration.undefined_value_symbol],
            #                                                       symbol,
            #                                                       s
            #                                                       )
            # self.__name_expression__ = helpers.__select_available_property__(name_expression,
            #                                                                  self.__name__,
            #                                                                  configuration.undefined_value_name)
            # self.__note__ = note
            # self.__symbol__ = helpers.__select_available_property__(symbol,
            #                                                         s,
            #                                                         configuration.undefined_value_symbol)
            # self.__symbolic_expression__ = helpers.__select_available_property__(symbolic_expression,
            #                                                                      se,
            #                                                                      configuration.undefined_value_symbolic_expression,
            #                                                                      )
            self.__numerical__ = helpers.__validate_numeric__(helpers.__select_available_property__(
                numerical,
                n,
                configuration.undefined_value_numerical))
            self.__print_style__ = print_style
            # self.__unit__ = helpers.__select_available_property__(first=unit,
            #                                                       second=u,
            #                                                       default=undefined_unit_default,
            #                                                       )

    @property
    def __type__(self):
        return "Value"

    def __add__(self, other):
        if configuration.setup.calculation_style == "numpsy":
            new = Value()
            name_variables = helpers.name_variables_generator(self, other)
            new.name_expression = (
                    "(" + name_variables[0] + "_plus_" + name_variables[1] + ")"
            )
            symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
            new.symbolic_expression = symbol_variables[0] + symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(self, other)
            new.numerical = numerical_variables[0] + numerical_variables[1]
            unit_variables = helpers.__unit_variables_generator__(self, other)
            new.unit = unit_variables[0] + unit_variables[1]
            return new
        if configuration.setup.calculation_style == "numpy":
            new = Value()
            # name_variables = helpers.name_variables_generator(self, other)
            # new.name_expression = (
            #         "(" + name_variables[0] + "_plus_" + name_variables[1] + ")"
            # )
            # symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
            # new.symbolic_expression = symbol_variables[0] + symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(self, other)
            new.numerical = numerical_variables[0] + numerical_variables[1]
            # unit_variables = helpers.__unit_variables_generator__(self, other)
            # new.unit = unit_variables[0] + unit_variables[1]
            return new

    def __neg__(self):
        if configuration.setup.calculation_style == "numpsy":
            new = Value()
            name_variable = helpers.name_variable_generator(self)
            new.name_expression = (
                    "minus" + name_variable
            )
            symbol_variable = helpers.symbolic_expression_variable_generator(self)
            new.symbolic_expression = - symbol_variable
            numerical_variable = helpers.numerical_variable_generator(self)
            new.numerical = - numerical_variable
            unit_variable = helpers.__unit_variable_generator__(self)
            new.unit = - unit_variable
            return new
        if configuration.setup.calculation_style == "numpy":
            new = Value()
            # name_variable = helpers.name_variable_generator(self)
            # new.name_expression = (
            #         "minus" + name_variable
            # )
            # symbol_variable = helpers.symbolic_expression_variable_generator(self)
            # new.symbolic_expression = - symbol_variable
            numerical_variable = helpers.numerical_variable_generator(self)
            new.numerical = - numerical_variable
            # unit_variable = helpers.__unit_variable_generator__(self)
            # new.unit = - unit_variable
            return new

    def __mul__(self, other):
        if configuration.setup.calculation_style == "numpsy":
            new = Value()
            name_variables = helpers.name_variables_generator(self, other)
            new.name_expression = (
                    "(" + name_variables[0] + "_times_" + name_variables[1] + ")"
            )
            new.name = new.name_expression
            symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
            new.symbolic_expression = symbol_variables[0] * symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(self, other)
            new.numerical = numerical_variables[0] * numerical_variables[1]
            unit_variables = helpers.__unit_variables_generator__(self, other)
            new.unit = unit_variables[0] * unit_variables[1]
            return new
        if configuration.setup.calculation_style == "numpy":
            new = Value()
            # name_variables = helpers.name_variables_generator(self, other)
            # new.name_expression = (
            #         "(" + name_variables[0] + "_times_" + name_variables[1] + ")"
            # )
            # new.name = new.name_expression
            # symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
            # new.symbolic_expression = symbol_variables[0] * symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(self, other)
            new.numerical = numerical_variables[0] * numerical_variables[1]
            # unit_variables = helpers.__unit_variables_generator__(self, other)
            # new.unit = unit_variables[0] * unit_variables[1]
            return new

    def __pow__(self, other):
        if configuration.setup.calculation_style == "numpsy":
            new = Value()
            name_variables = helpers.name_variables_generator(self, other)
            new.name_expression = (
                    "(" + name_variables[0] + "_power_" + name_variables[1] + ")"
            )
            symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
            new.symbolic_expression = symbol_variables[0] ** symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(self, other)
            new.numerical = numerical_variables[0] ** numerical_variables[1]
            unit_variables = helpers.__unit_variables_generator__(self, other)
            new.unit = unit_variables[0] ** unit_variables[1]
            return new
        if configuration.setup.calculation_style == "numpy":
            new = Value()
            # name_variables = helpers.name_variables_generator(self, other)
            # new.name_expression = (
            #         "(" + name_variables[0] + "_power_" + name_variables[1] + ")"
            # )
            # symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
            # new.symbolic_expression = symbol_variables[0] ** symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(self, other)
            new.numerical = numerical_variables[0] ** numerical_variables[1]
            # unit_variables = helpers.__unit_variables_generator__(self, other)
            # new.unit = unit_variables[0] ** unit_variables[1]
            return new

    def __sub__(self, other):
        if configuration.setup.calculation_style == "numpsy":
            new = Value()
            name_variables = helpers.name_variables_generator(self, other)
            new.name_expression = (
                    "(" + name_variables[0] + "_minus_" + name_variables[1] + ")"
            )
            symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
            new.symbolic_expression = symbol_variables[0] - symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(self, other)
            new.numerical = numerical_variables[0] - numerical_variables[1]
            unit_variables = helpers.__unit_variables_generator__(self, other)
            new.unit = unit_variables[0] - unit_variables[1]
            return new
        if configuration.setup.calculation_style == "numpy":
            new = Value()
            # name_variables = helpers.name_variables_generator(self, other)
            # new.name_expression = (
            #         "(" + name_variables[0] + "_minus_" + name_variables[1] + ")"
            # )
            # symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
            # new.symbolic_expression = symbol_variables[0] - symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(self, other)
            new.numerical = numerical_variables[0] - numerical_variables[1]
            # unit_variables = helpers.__unit_variables_generator__(self, other)
            # new.unit = unit_variables[0] - unit_variables[1]
            return new

    def __radd__(self, other):
        if configuration.setup.calculation_style == "numpsy":
            new = Value()
            name_variables = helpers.name_variables_generator(other, self)
            new.name_expression = (
                    "(" + name_variables[0] + "_plus_" + name_variables[1] + ")"
            )
            symbol_variables = helpers.symbolic_expression_variables_generator(other, self)
            new.symbolic_expression = symbol_variables[0] + symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(other, self)
            new.numerical = numerical_variables[0] + numerical_variables[1]
            unit_variables = helpers.__unit_variables_generator__(other, self)
            new.unit = unit_variables[0] + unit_variables[1]
            return new
        if configuration.setup.calculation_style == "numpy":
            new = Value()
            # name_variables = helpers.name_variables_generator(other, self)
            # new.name_expression = (
            #         "(" + name_variables[0] + "_plus_" + name_variables[1] + ")"
            # )
            # symbol_variables = helpers.symbolic_expression_variables_generator(other, self)
            # new.symbolic_expression = symbol_variables[0] + symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(other, self)
            new.numerical = numerical_variables[0] + numerical_variables[1]
            # unit_variables = helpers.__unit_variables_generator__(other, self)
            # new.unit = unit_variables[0] + unit_variables[1]
            return new

    def __rmul__(self, other):
        if configuration.setup.calculation_style == "numpsy":
            new = Value()
            name_variables = helpers.name_variables_generator(other, self)
            new.name_expression = (
                    "(" + name_variables[0] + "_times_" + name_variables[1] + ")"
            )
            symbol_variables = helpers.symbolic_expression_variables_generator(other, self)
            new.symbolic_expression = symbol_variables[0] * symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(other, self)
            new.numerical = numerical_variables[0] * numerical_variables[1]
            unit_variables = helpers.__unit_variables_generator__(other, self)
            new.unit = unit_variables[0] * unit_variables[1]
            return new
        if configuration.setup.calculation_style == "numpy":
            new = Value()
            # name_variables = helpers.name_variables_generator(other, self)
            # new.name_expression = (
            #         "(" + name_variables[0] + "_times_" + name_variables[1] + ")"
            # )
            # symbol_variables = helpers.symbolic_expression_variables_generator(other, self)
            # new.symbolic_expression = symbol_variables[0] * symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(other, self)
            new.numerical = numerical_variables[0] * numerical_variables[1]
            # unit_variables = helpers.__unit_variables_generator__(other, self)
            # new.unit = unit_variables[0] * unit_variables[1]
            return new

    def __rsub__(self, other):
        if configuration.setup.calculation_style == "numpsy":
            new = Value()
            name_variables = helpers.name_variables_generator(other, self)
            new.name_expression = (
                    "(" + name_variables[0] + "_minus_" + name_variables[1] + ")"
            )
            symbol_variables = helpers.symbolic_expression_variables_generator(other, self)
            new.symbolic_expression = symbol_variables[0] - symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(other, self)
            new.numerical = numerical_variables[0] - numerical_variables[1]
            unit_variables = helpers.__unit_variables_generator__(other, self)
            new.unit = unit_variables[0] - unit_variables[1]
            return new
        if configuration.setup.calculation_style == "numpy":
            new = Value()
            # name_variables = helpers.name_variables_generator(other, self)
            # new.name_expression = (
            #         "(" + name_variables[0] + "_minus_" + name_variables[1] + ")"
            # )
            # symbol_variables = helpers.symbolic_expression_variables_generator(other, self)
            # new.symbolic_expression = symbol_variables[0] - symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(other, self)
            new.numerical = numerical_variables[0] - numerical_variables[1]
            # unit_variables = helpers.__unit_variables_generator__(other, self)
            # new.unit = unit_variables[0] - unit_variables[1]
            return new

    def __rtruediv__(self, other):
        if configuration.setup.calculation_style == "numpsy":
            new = Value()
            name_variables = helpers.name_variables_generator(other, self)
            new.name_expression = (
                    "(" + name_variables[0] + ")_per_(" + name_variables[1] + ")"
            )
            symbol_variables = helpers.symbolic_expression_variables_generator(other, self)
            new.symbolic_expression = symbol_variables[0] / symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(other, self)
            new.numerical = numerical_variables[0] / numerical_variables[1]
            unit_variables = helpers.__unit_variables_generator__(other, self)
            new.unit = unit_variables[0] / unit_variables[1]
            return new
        if configuration.setup.calculation_style == "numpy":
            new = Value()
            name_variables = helpers.name_variables_generator(other, self)
            # new.name_expression = (
            #         "(" + name_variables[0] + ")_per_(" + name_variables[1] + ")"
            # )
            # symbol_variables = helpers.symbolic_expression_variables_generator(other, self)
            # new.symbolic_expression = symbol_variables[0] / symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(other, self)
            new.numerical = numerical_variables[0] / numerical_variables[1]
            # unit_variables = helpers.__unit_variables_generator__(other, self)
            # new.unit = unit_variables[0] / unit_variables[1]
            return new

    def __truediv__(self, other):
        if configuration.setup.calculation_style == "numpsy":
            new = Value()
            name_variables = helpers.name_variables_generator(self, other)
            new.name_expression = (
                    "(" + name_variables[0] + ")_per_(" + name_variables[1] + ")"
            )
            symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
            new.symbolic_expression = symbol_variables[0] / symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(self, other)
            new.numerical = numerical_variables[0] / numerical_variables[1]
            unit_variables = helpers.__unit_variables_generator__(self, other)
            new.unit = unit_variables[0] / unit_variables[1]
            return new
        if configuration.setup.calculation_style == "numpy":
            new = Value()
            # name_variables = helpers.name_variables_generator(self, other)
            # new.name_expression = (
            #         "(" + name_variables[0] + ")_per_(" + name_variables[1] + ")"
            # )
            # symbol_variables = helpers.symbolic_expression_variables_generator(self, other)
            # new.symbolic_expression = symbol_variables[0] / symbol_variables[1]
            numerical_variables = helpers.numerical_variables_generator(self, other)
            new.numerical = numerical_variables[0] / numerical_variables[1]
            # unit_variables = helpers.__unit_variables_generator__(self, other)
            # new.unit = unit_variables[0] / unit_variables[1]
            return new

    def lambdify_symbolic_expression(self,
                                     lambdify_string=configuration.undefined_unit_symbol):
        """
        This method uses sympy lambdify on the symbolic expression based on the lambdify_symbol term.
        This is an issue when using the numpy only calculation mode, as the lambdification process would not effectively work.
        """
        if configuration.setup.calculation_style == "numpsy":
            lambdify_symbol = sy.Symbol(lambdify_string)
            return sy.lambdify(lambdify_symbol,
                               self.symbolic_expression,
                               "numpy")
        if configuration.setup.calculation_style == "numpy":
            lambdify_symbol = sy.Symbol(lambdify_string)
            return sy.lambdify(lambdify_symbol,
                               self.symbolic_expression,
                               "numpy")
    @property
    def numerical(self):
        """Return unit numerical value shorthand"""
        return self.__numerical__

    @numerical.setter
    def numerical(self, value):
        self.__numerical__ = value

    @property
    def symbol(self):
        """Return unit symbol value shorthand"""
        if isinstance(self.__symbol__, str):
            return sy.Symbol(self.__symbol__)
        else:
            return self.__symbol__

    @symbol.setter
    def symbol(self, value):
        self.__symbol__ = value

    @property
    def symbolic_expression(self):
        """Return symbolic expression consisting of an equivalent mathemtical representation to symbol"""
        return self.__symbolic_expression__

    @symbolic_expression.setter
    def symbolic_expression(self, value):
        """Set symbolic expression consisting of an equivalent mathemtical representation to symbol."""
        self.__symbolic_expression__ = value

    @property
    def unit(self):
        """Return unit symbol unit shorthand"""
        return self.__unit__

    @unit.setter
    def unit(self, value):
        self.__unit__ = value

    # Shorthands
    u = unit
    s = symbol
    n = numerical
    se = symbolic_expression


class Constant(Value):
    @Value.numerical.setter
    def numerical(self, value):
        print(
            Warning(
                "Constant cannot be mutated. You cannot set any attribute value. Instantiate a new variable."
            )
        )

    @Value.symbol.setter
    def symbol(self, value):
        print(
            Warning(
                "Constant cannot be mutated. You cannot set any attribute value. Instantiate a new variable."
            )
        )

    @Value.unit.setter
    def unit(self, value):
        print(
            Warning(
                "Constant cannot be mutated. You cannot set any attribute value. Instantiate a new variable."
            )
        )

    @Value.symbolic_expression.setter
    def symbolic_expression(self, value):
        """Set symbolic expression shorthand shorthand"""
        print(
            Warning(
                "Constant cannot be mutated. You cannot set any attribute value. Instantiate a new variable."
            )
        )

    # Shorthands
    u = unit
    s = symbol
    n = numerical
    se = symbolic_expression


class Variable(Value):
    pass


# Shorthand classes
C = Constant
V = Variable
U = Unit



import pandas as pd
import scipy as sp
from . import core
from . import units


class Constants:
    # TODO connect to scipy library
    # TODO import from Pandas CSV
    def __init__(self):
        self.units = units.Units().data

    @property
    def data(self):
        "Symbols from Sympy"
        data_frame = pd.DataFrame(
            {
                "impedance_free_space": core.Constant(
                    name="impedance_free_space",
                    symbol="\eta_0",
                    numerical=377,
                    unit=self.units.ohm,
                ),
                "permeability_relative_vaccum": core.Constant(
                    name="permeability_relative_vaccum",
                    symbol="\mu_{r0}",
                    numerical=1,
                    unit=self.units.Henry / self.units.meter,
                ),
                "permittivity_relative_intrinsic_silicon": core.Constant(
                    name="permeability_relative_intrinsic_silicon",
                    symbol="\epsilon_{Si}",
                    numerical=11.7,
                    unit=self.units.Farad / self.units.meter,
                ),
                "permittivity_relative_silicon_dioxide": core.Constant(
                    name="permeability_relative_silicon_dioxide",
                    symbol="\epsilon_{SiO2}",
                    numerical=3.9,
                    unit=self.units.Farad / self.units.meter,
                ),
                "permittivity_relative_vaccum": core.Constant(
                    name="permeability_relative_vaccum",
                    symbol="\epsilon_{r0}",
                    numerical=1,
                    unit=self.units.Farad / self.units.meter,
                ),
                "permittivity_vaccum": core.Constant(
                    name="permittivity_vaccum",
                    symbol="\epsilon_0",
                    numerical=8.8541878128e-12,
                    unit=self.units.Farad / self.units.meter,
                ),
                "pi": core.Constant(
                    name="pi",
                    symbol="\pi",
                    numerical=sp.pi
                ),
                "speed_of_light": core.Constant(
                    name="speed_of_light",
                    symbol="c",
                    numerical=299792458,
                    unit=self.units.meter / self.units.second,
                ),
            }, index=[0])

        data_frame["c"] = data_frame.speed_of_light
        data_frame["e_0"] = data_frame.permittivity_vaccum
        data_frame["e_r_0"] = data_frame.permittivity_relative_vaccum
        data_frame["e_r_si"] = data_frame.permittivity_relative_intrinsic_silicon
        data_frame["e_r_sio2"] = data_frame.permittivity_relative_silicon_dioxide
        data_frame["eta_0"] = data_frame.impedance_free_space
        data_frame["mu_r_0"] = data_frame.permeability_relative_vaccum

        return data_frame.iloc[0]

# TODO potentially rename to constants library
c = Constants().data



import copy
from . import configuration



def __repr__(instance):
    """Main pretty printing instance representation method.

    The goal is that when it prints in IPython it shows its parameters as rows,
    kind of like Pandas tables, except that in this case we would have both
    symbol, symbolic_expression, numerical, unit and name as the default prints.
    Other special methods can print the full numpsy instance.

    Note that because Unit is a NumpSy class, we then need to make it embeddable into a
    Variable representation, and both representations should be adaptable. So this representation
    is hierarchical.

    It also aims to enhance the functionality available via the sympy printers
    """
    output = instance.data
    return instance.__class__.__name__ + str(output)


def markdownify(instance):
    if hasattr(instance, "print_style") and bool(instance.print_style):
        if instance.print_style == configuration.available_print_styles["numpsy"]:
            if instance.__class__.__name__ == "Unit":
                display_columns = ["name", "symbol", "symbolic_expression"]
            else:
                display_columns = configuration.setup.value_printer_columns
            markdown_instance = copy.deepcopy(instance)
            return markdown_instance.data.loc[:, display_columns].T.to_markdown()

        if instance.print_style == configuration.available_print_styles["numpy"]:
            if instance.__class__.__name__ == "Unit":
                display_columns = ["symbolic_expression"]
                markdown_instance = copy.deepcopy(instance)
                return markdown_instance.data.loc[:, display_columns].T.to_markdown()
            else:
                display_columns = ["numerical"]
                markdown_instance = copy.deepcopy(instance)
                return markdown_instance.data.loc[:, display_columns].T.to_markdown()

        if instance.print_style == configuration.available_print_styles["sympy"]:
            if instance.__class__.__name__ == "Unit":
                display_columns = ["symbol", "symbolic_expression"]
            else:
                display_columns = ["symbol", "symbolic_expression"]
            markdown_instance = copy.deepcopy(instance)
            return markdown_instance.data.loc[:, display_columns].T.to_markdown()
    else:
        if configuration.setup.print_style == configuration.available_print_styles["numpsy"]:
            if instance.__class__.__name__ == "Unit":
                display_columns = ["name", "symbol", "symbolic_expression"]
            else:
                display_columns = configuration.setup.value_printer_columns
            markdown_instance = copy.deepcopy(instance)
            return markdown_instance.data.loc[:, display_columns].T.to_markdown()

        if configuration.setup.print_style == configuration.available_print_styles["numpy"]:
            if instance.__class__.__name__ == "Unit":
                display_columns = ["symbolic_expression"]
                markdown_instance = copy.deepcopy(instance)
                return markdown_instance.data.loc[:, display_columns].T.to_markdown()
            else:
                display_columns = ["numerical"]
                markdown_instance = copy.deepcopy(instance)
                return markdown_instance.data.loc[:, display_columns].T.to_markdown()

        if configuration.setup.print_style == configuration.available_print_styles["sympy"]:
            if instance.__class__.__name__ == "Unit":
                display_columns = ["symbol", "symbolic_expression"]
            else:
                display_columns = ["symbol", "symbolic_expression"]
            markdown_instance = copy.deepcopy(instance)
            return markdown_instance.data.loc[:, display_columns].T.to_markdown()

def _repr_markdown_(display_dataframe):
    return markdownify(display_dataframe)



import numpy as np
import sympy as sy

undefined_note = ""

undefined_unit_name = "undefined"
undefined_unit_symbol = "Ø"
undefined_unit_symbolic_expression = sy.Symbol(undefined_unit_symbol)

undefined_value_name = "undefined"
undefined_value_symbol = "Ø"
undefined_value_symbolic_expression = sy.Symbol(undefined_value_symbol)

undefined_value_numerical = np.array([])



available_calculation_styles = {
    "numpsy": "numpsy",
    "numpy": "numpy",
    "sympy": "sympy"
}
undefined_calculation_style = None
default_calculation_style = available_calculation_styles["numpsy"]

available_matrix_operation_mode = {
    "vectorial": "vectorial",
}
default_matrix_operation_mode = available_matrix_operation_mode["vectorial"]

available_print_styles = {
    "numpsy": "numpsy",
    "numpy": "numpy",
    "sympy": "sympy"
}
undefined_print_style = None
default_print_style = available_print_styles["numpsy"]
default_debug_mode = False

default_numpsy_value_printer_columns = ["name", "symbol", "symbolic_expression", "numerical", "unit", "note"]

class ConfigurationSetup:

    def __init__(self,
                 calculation_style=default_calculation_style,
                 debug_mode=default_debug_mode,
                 matrix_operation_mode=default_matrix_operation_mode,
                 print_style=default_print_style,
                 value_printer_columns=default_numpsy_value_printer_columns
                 ):
        self.__calculation_style__ = calculation_style
        self.__debug_mode__ = debug_mode
        self.__matrix_operation_mode__ = matrix_operation_mode
        self.__print_style__ = print_style
        self.__value_printer_columns__ = value_printer_columns

    @property
    def debug_mode(self):
        """Return name string"""
        return self.__debug_mode__

    @debug_mode.setter
    def debug_mode(self, value):
        self.__debug_mode__ = value

    @property
    def calculation_style(self):
        """
        Part of the complexity is that there are some circumstances where for a local operation, a calculation_style may be desired, but not necessarily for a global case.
        """
        return self.__calculation_style__

    @calculation_style.setter
    def calculation_style(self, value):
        self.__calculation_style__ = value

    @property
    def matrix_operation_mode(self):
        return self.__matrix_operation_mode__

    @matrix_operation_mode.setter
    def matrix_operation_mode(self, value):
        self.__matrix_operation_mode__ = value

    @property
    def print_style(self):
        """Return name string"""
        return self.__print_style__

    @print_style.setter
    def print_style(self, value):
        self.__print_style__ = value

    @property
    def value_printer_columns(self):
        """Return name string"""
        return self.__value_printer_columns__

    @value_printer_columns.setter
    def value_printer_columns(self, value):
        self.__value_printer_columns__ = value

setup = ConfigurationSetup()



import numpy as np
import sympy as sy
import sympy.functions.special.elliptic_integrals
import scipy as sp
import scipy.special
from . import configuration
from . import core
from . import helpers


def __variable_compatibility_check__(input):
    if hasattr(input, "__parent_class__"):
        if input.__parent_class__ == "Value":
            new = input.__class__()
        else:
            new = core.Variable()
    else:
        new = core.Variable()
    return new

def abs(instance=core.Variable()):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.abs(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        new.symbolic_expression = sy.Abs(instance_parameters["symbolic"])
        new.name_expression = "abs(" + instance_parameters["name"] + ")"
        return new
    if configuration.setup.calculation_style == "numpy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.abs(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        # new.symbolic_expression = sy.Abs(instance_parameters["symbolic"])
        # new.name_expression = "abs(" + instance_parameters["name"] + ")"
        return new

def binomial_coefficient(instance=core.Variable(), instance_2=core.Variable()):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        new_2 = __variable_compatibility_check__(instance_2)
        instance_parameters = helpers.full_variable_generator(instance)
        instance_parameters_2 = helpers.full_variable_generator(instance_2)
        new.numerical = sp.special.binom(instance_parameters["numerical"], instance_parameters_2["numerical"])
        # TODO operate on both symbolic variables
        new.symbolic_expression = sy.binomial(instance_parameters["symbolic"], instance_parameters_2["symbolic"])
        new.name_expression = "binomial(" + instance_parameters["name"] + "," + instance_parameters_2["name"] + ")"
        return new
    if configuration.setup.calculation_style == "numpy":
        new = __variable_compatibility_check__(instance)
        new_2 = __variable_compatibility_check__(instance_2)
        instance_parameters = helpers.full_variable_generator(instance)
        instance_parameters_2 = helpers.full_variable_generator(instance_2)
        new.numerical = sp.special.binom(instance_parameters["numerical"], instance_parameters_2["numerical"])
        # TODO operate on both symbolic variables
        # new.symbolic_expression = sy.binomial(instance_parameters["symbolic"], instance_parameters_2["symbolic"])
        # new.name_expression = "binomial(" + instance_parameters["name"] + "," + instance_parameters_2["name"] + ")"
        return new

def ciel(instance=core.Variable()):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.ceil(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        new.symbolic_expression = sy.ceiling(instance_parameters["symbolic"])
        new.name_expression = "ceil(" + instance_parameters["name"] + ")"
        return new
    if configuration.setup.calculation_style == "numpy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.ceil(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        # new.symbolic_expression = sy.ceiling(instance_parameters["symbolic"])
        # new.name_expression = "ceil(" + instance_parameters["name"] + ")"
        return new

def complete_elliptical_integral_first_kind(instance=core.Variable()):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = sp.special.ellipk(instance_parameters["numerical"])
        new.symbolic_expression = sy.functions.special.elliptic_integrals.elliptic_k(
            instance_parameters["symbolic"]
        )
        new.name_expression = (
            "complete_elliptical_integral_first_kind(" + instance_parameters["name"] + ")"
        )
        # TODO automate this
        new.unit.name = "K(" + instance_parameters["unit"].name + ")"
        # TODO get symbolic parameters
        new.unit.symbolic_expression = sy.functions.special.elliptic_integrals.elliptic_k(
            instance_parameters["unit"].symbolic_expression
        )
        return new
    if configuration.setup.calculation_style == "numpy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = sp.special.ellipk(instance_parameters["numerical"])
        # new.symbolic_expression = sy.functions.special.elliptic_integrals.elliptic_k(
        #     instance_parameters["symbolic"]
        # )
        # new.name_expression = (
        #         "complete_elliptical_integral_first_kind(" + instance_parameters["name"] + ")"
        # )
        # # TODO automate this
        # new.unit.name = "K(" + instance_parameters["unit"].name + ")"
        # # TODO get symbolic parameters
        # new.unit.symbolic_expression = sy.functions.special.elliptic_integrals.elliptic_k(
        #     instance_parameters["unit"].symbolic_expression
        # )
        return new

def exp(instance=core.Variable()):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.exp(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        new.symbolic_expression = sy.exp(instance_parameters["symbolic"])
        new.name_expression = "exp(" + instance_parameters["name"] + ")"
        return new
    if configuration.setup.calculation_style == "numpy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.exp(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        # new.symbolic_expression = sy.exp(instance_parameters["symbolic"])
        # new.name_expression = "exp(" + instance_parameters["name"] + ")"
        return new


def log(instance=core.Variable()):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.log(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        new.symbolic_expression = sy.log(instance_parameters["symbolic"])
        new.name_expression = "ln(" + instance_parameters["name"] + ")"
        return new
    if configuration.setup.calculation_style == "numpy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.log(instance_parameters["numerical"])
        # # TODO operate on both symbolic variables
        # new.symbolic_expression = sy.log(instance_parameters["symbolic"])
        # new.name_expression = "ln(" + instance_parameters["name"] + ")"
        return new

def log2(instance=core.Variable()):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.log2(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        new.symbolic_expression = sy.log(instance_parameters["symbolic"], 2)
        new.name_expression = "log2(" + instance_parameters["name"] + ")"
        return new
    if configuration.setup.calculation_style == "numpy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.log2(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        # new.symbolic_expression = sy.log(instance_parameters["symbolic"], 2)
        # new.name_expression = "log2(" + instance_parameters["name"] + ")"
        return new

def log10(instance=core.Variable()):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.log10(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        new.symbolic_expression = sy.log(instance_parameters["symbolic"], 10)
        new.name_expression = "exp(" + instance_parameters["name"] + ")"
        return new
    if configuration.setup.calculation_style == "numpy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.log10(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        # new.symbolic_expression = sy.log(instance_parameters["symbolic"], 10)
        # new.name_expression = "exp(" + instance_parameters["name"] + ")"
        return new

def prod(instance=core.Variable(),
         sum_variable_string="i",
         minimum_range_amount=0,
         maximum_range_amount=None
         ):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.prod(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        new.symbolic_expression = sy.Product(instance_parameters["symbolic"], (sy.Symbol(sum_variable_string),
                                                                               minimum_range_amount,
                                                                               maximum_range_amount))
        new.name_expression = "prod(" + instance_parameters["name"] + ")"
        return new
    if configuration.setup.calculation_style == "numpy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.prod(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        # new.symbolic_expression = sy.Product(instance_parameters["symbolic"], (sy.Symbol(sum_variable_string),
        #                                                                        minimum_range_amount,
        #                                                                        maximum_range_amount))
        # new.name_expression = "prod(" + instance_parameters["name"] + ")"
        return new

def repeat(instance=core.Variable(), repeats=1):
    if configuration.setup.calculation_style == "numpsy":
        # TODO unconvinced about the symbolic implementation of this.
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        # print(instance_parameters["numerical"])
        new.numerical = np.repeat(instance_parameters["numerical"], repeats)
        # TODO operate on both symbolic variables
        new.symbolic_expression = instance_parameters["symbolic"]
        new.name_expression = "repeat(" + instance_parameters["name"] + ")"
        return new
    if configuration.setup.calculation_style == "numpy":
        # TODO unconvinced about the symbolic implementation of this.
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        # print(instance_parameters["numerical"])
        new.numerical = np.repeat(instance_parameters["numerical"], repeats)
        # # TODO operate on both symbolic variables
        # new.symbolic_expression = instance_parameters["symbolic"]
        # new.name_expression = "repeat(" + instance_parameters["name"] + ")"
        return new

def sqrt(instance=core.Variable()):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.sqrt(instance_parameters["numerical"])
        new.symbolic_expression = sy.sqrt(instance_parameters["symbolic"])
        new.name_expression = "square_root(" + instance_parameters["name"] + ")"
        # TODO automate this
        new.unit.name = "square_root(" + instance_parameters["unit"].name + ")"
        # TODO get symbolic parameters
        new.unit.symbolic_expression = sy.sqrt(
            instance_parameters["unit"].symbolic_expression
        )
        return new
    if configuration.setup.calculation_style == "numpy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.sqrt(instance_parameters["numerical"])
        # new.symbolic_expression = sy.sqrt(instance_parameters["symbolic"])
        # new.name_expression = "square_root(" + instance_parameters["name"] + ")"
        # # TODO automate this
        # new.unit.name = "square_root(" + instance_parameters["unit"].name + ")"
        # # TODO get symbolic parameters
        # new.unit.symbolic_expression = sy.sqrt(
        #     instance_parameters["unit"].symbolic_expression
        # )
        return new

def sinh(instance=core.Variable()):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.sinh(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        new.symbolic_expression = sy.sinh(instance_parameters["symbolic"])
        new.name_expression = "sinh(" + instance_parameters["name"] + ")"
        return new
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.sinh(instance_parameters["numerical"])
        # # TODO operate on both symbolic variables
        # new.symbolic_expression = sy.sinh(instance_parameters["symbolic"])
        # new.name_expression = "sinh(" + instance_parameters["name"] + ")"
        return new

def sum(instance=core.Variable(),
        sum_variable_string="i",
        minimum_range_amount=0,
        maximum_range_amount=0,
        axis=0):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.sum(instance_parameters["numerical"], axis=axis)
        # TODO operate on both symbolic variables
        new.symbolic_expression = sy.Sum(instance_parameters["symbolic"], (sy.Symbol(sum_variable_string),
                                                                           minimum_range_amount,
                                                                           maximum_range_amount))
        new.name_expression = "sum(" + instance_parameters["name"] + ")"
        return new
    if configuration.setup.calculation_style == "numpy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.sum(instance_parameters["numerical"], axis=axis)
        # TODO operate on both symbolic variables
        # new.symbolic_expression = sy.Sum(instance_parameters["symbolic"], (sy.Symbol(sum_variable_string),
        #                                                                    minimum_range_amount,
        #                                                                    maximum_range_amount))
        # new.name_expression = "sum(" + instance_parameters["name"] + ")"
        return new

def tanh(instance=core.Variable()):
    if configuration.setup.calculation_style == "numpsy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.tanh(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        new.symbolic_expression = sy.tanh(instance_parameters["symbolic"])
        new.name_expression = "tanh(" + instance_parameters["name"] + ")"
        return new
    if configuration.setup.calculation_style == "numpy":
        new = __variable_compatibility_check__(instance)
        instance_parameters = helpers.full_variable_generator(instance)
        new.numerical = np.tanh(instance_parameters["numerical"])
        # TODO operate on both symbolic variables
        # new.symbolic_expression = sy.tanh(instance_parameters["symbolic"])
        # new.name_expression = "tanh(" + instance_parameters["name"] + ")"
        return new



e = exp



import pandas as pd
from . import core


class Units:
    # TODO append method
    def __init__(self):
        pass

    @property
    def data(self):
        "Full dataframe"
        return pd.DataFrame(
            {
                "Hertz": core.Unit(name="Hertz", symbol="Hz"),
                "Henry": core.Unit(name="Henry", symbol="H"),
                "Farad": core.Unit(name="Farad", symbol="F"),
                "meter": core.Unit(name="meter", symbol="m"),
                "ohm": core.Unit(name="ohm", symbol="\Omega"),
                "second": core.Unit(name="second", symbol="s"),
                "ratio": core.Unit(name="ratio", symbol=""),
            },
            index=[0],
        ).iloc[0]

    @data.setter
    def data(self, unit):
        raise NotImplementedError("TODO units appender")


u = Units().data



#!/usr/bin/env python
import setuptools

# # read the contents of your README file
# from pathlib import Path
# this_directory = Path(__file__).parent
# long_description = (this_directory / "README_RAW.md").read_text()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="numpsy",
    version="0.0.11",
    description="NumpSy - Integrated NumPy, SymPy, SciPy and Pandas with unit management for scientific programming with IPython.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/daquintero/numpsy",
    author="Dario Quintero",
    author_email="darioaquintero@gmail.com",
    packages=setuptools.find_packages(),
    include_package_data=True,
    # install_requires=[
    #     "ipython",
    #     "numpy",
    #     "pandas",
    #     "sympy",
    #     # "scipy",
    #     "cython",
    #     "tabulate",
    # ],
    # TODO sort out requirements outside of Anaconda3
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
